
import java.util.ArrayList;
import java.util.Arrays;

/**
 * This class will handle various interactions with a byte array.
 *
 * @author Roger Lacroix, Capitalware Inc.
 * @version 1.0.0
 * @license Apache 2 License
 */
public class ByteArray
{
   private static final char[] HEX_ARRAY = "0123456789ABCDEF".toCharArray();
   private byte[] bArray;

   /**
    * The constructor
    *
    * @param bArray use the byte array passed in.
    */
   public ByteArray(byte[] bArray)
   {
      super();
      this.bArray = bArray;
   }

   /**
    * Returns the byte value at the specified index.
    * @param index within the byte array to select the byte.
    * @return a byte
    */
   public byte byteAt(int index)
   {
      if (index >= bArray.length)
         throw new IndexOutOfBoundsException("Index is larger than length of byte array.");

      return bArray[index];
   }

   /**
    * Returns the character value at the specified index.
    * @param index within the byte array to select the byte.
    * @return a character
    */
   public char charAt(int index)
   {
      if (index >= bArray.length)
         throw new IndexOutOfBoundsException("Index is larger than length of byte array.");

      return (char)bArray[index];
   }

   /**
    * Concatenates the specified byte array to the end of this byte array.
    * @param suffixArray inputed byte array for concatenation.
    * @return updated byte array with inputed byte array added to the end
    */
   public byte[] concat(byte[] suffixArray)
   {
      byte[] bOutput = new byte[bArray.length + suffixArray.length];

      for (int i=0; i < bArray.length; i++)
         bOutput[i] = bArray[i];

      for (int i=bArray.length, j=0; j < suffixArray.length; i++, j++)
         bOutput[i] = suffixArray[j];

      return bOutput;
   }

   /**
    * Convert an integer to Big Endian format.
    * @return
    */
   public int convertToInt_BE()
   {
      int xx =  (bArray[0] << 24) +
               ((bArray[1] & 0xFF) << 16) +
               ((bArray[2] & 0xFF) << 8) +
                (bArray[3] & 0xFF);
            
      return xx;
   }
   
   /**
    * Tests if this byte array ends with the specified suffix.
    * @param suffixArray inputed byte array
    * @return true if the byte array ends with the inputed byte array, otherwise false
    */
   public boolean endsWith(byte[] suffixArray)
   {
      boolean flag = true;

      if (bArray.length < suffixArray.length)
         return false;

      int s = bArray.length - suffixArray.length;
      for (int i=s, j=0; j < suffixArray.length; i++, j++)
      {
         if (bArray[i] != suffixArray[j])
         {
            flag = false;
            break;
         }
      }

      return flag;
   }

   /**
    * Value equality for byte arrays.
    *
    * @param other object to be compared against
    * @return true if both byte arrays exactly match, otherwise false
    */
   public boolean equals(Object other)
   {
      if (other instanceof ByteArray)
      {
         ByteArray ba = (ByteArray) other;
         return equals(bArray, bArray.length, ba.getBytes(), ba.getLength(), 0);
      }

      return false;
   }

   /**
    * Value equality for byte arrays.
    *
    * @param other object to be compared against
    * @param offset offset to start the comparison at.
    * @return true if both byte arrays exactly match, otherwise false
    */
   public boolean equals(Object other, int offset)
   {
      if (other instanceof ByteArray)
      {
         ByteArray ba = (ByteArray) other;
         return equals(bArray, bArray.length, ba.getBytes(), (ba.getLength() - offset), offset);
      }

      return false;
   }

   /**
    * Compare two byte arrays using value equality. Two byte arrays are equal if
    * their length is identical and their contents are identical.
    * @param b1  byte array #1 to be checked
    * @param b1Length length of of byte array #1
    * @param b2  byte array #2 to be checked
    * @param b2Length length of of byte array #2
    * @param b2Offset starting position within byte array #2
    * @return true if both byte arrays exactly match, otherwise false
    */
   private boolean equals(byte[] b1, int b1Length, byte[] b2, int b2Length, int b2Offset)
   {
      if (b1Length != b2Length)
         return false;

      for (int i = 0; i < b1Length; i++)
      {
         if (b1[i] != b2[i + b2Offset])
            return false;
      }
      return true;
   }

   /**
    * Return the entire byte array
    *
    * @return bArray
    */
   public byte[] getBytes()
   {
      return bArray;
   }

   /**
    * Returns a byte array that is a sub byte array of this byte array. The sub byte array begins with
    * the byte at the specified index and extends to the end of this byte array.
    * @param beginIndex starting index for the selection
    * @return sub byte array
    * @throws IndexOutOfBoundsException
    */
   public byte[] getBytes(int beginIndex) throws IndexOutOfBoundsException
   {
      if (beginIndex >= bArray.length)
         throw new IndexOutOfBoundsException("beginIndex is larger than length of byte array.");

      int newLen = bArray.length - beginIndex;
      byte[] bOutput = new byte[newLen];

      for (int i = 0; i < newLen; i++)
      {
         bOutput[i] = bArray[i + beginIndex];
      }

      return bOutput;
   }

   /**
    * Returns a byte array that is a sub byte array of this byte array. The sub byte array begins at
    * the specified beginIndex and extends to the byte at index endIndex - 1. Thus the length of the
    * sub byte array is endIndex-beginIndex.
    * @param beginIndex starting index for the selection
    * @param endIndex ending index for the selection
    * @return sub byte array
    * @throws IndexOutOfBoundsException
    */
   public byte[] getBytes(int beginIndex, int endIndex) throws IndexOutOfBoundsException
   {
      if (beginIndex >= endIndex)
         throw new IndexOutOfBoundsException("beginIndex is greater than or equal to endIndex.");
      else if (beginIndex >= bArray.length)
         throw new IndexOutOfBoundsException("beginIndex is larger than length of byte array.");
      else if (endIndex > bArray.length)
         throw new IndexOutOfBoundsException("endIndex is larger than length of byte array.");

      int newLen = endIndex - beginIndex;
      byte[] bOutput = new byte[newLen];

      for (int i = 0; i < newLen; i++)
      {
         bOutput[i] = bArray[i + beginIndex];
      }

      return bOutput;
   }

   /**
    * Returns the length of the byte array.
    *
    * @return length of byte array
    */
   public int getLength()
   {
      return bArray.length;
   }

   /**
    * Returns the index within this larger byte array of the first occurrence of
    * the specified byte array. Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the byte array in the larger
    *         byte array or -1 if not found.
    */
   public int indexOf(byte[] searchFor)
   {
      return indexOf(searchFor, 0);
   }

   /**
    * Returns the index within this larger byte array of the first occurrence of
    * the specified byte array, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the byte array in the larger
    *         byte array or -1 if not found.
    */
   public int indexOf(byte[] searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;
      int searchForLen = searchFor.length;

      // Starting point must be zero or higher.
      if (fromIndex < 0)
         return -1;
      // Is the starting position within the byte array?
      else if (fromIndex >= arrayLen)
         return -1;

      // Is the searchFor byte array larger than the byte array to be search?
      if (searchForLen > arrayLen)
         return -1;

      for (int i = fromIndex; i < arrayLen - searchForLen + 1; i++)
      {
         boolean found = true;
         for (int j = 0; j < searchForLen; j++)
         {
            if (bArray[i + j] != searchFor[j])
            {
               found = false;
               break;
            }
         }
         if (found)
            return i;
      }

      return -1;
   }

   /**
    * Returns the index within this byte array of the first occurrence of the
    * specified byte.  Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the byte in the larger byte
    *         array or -1 if not found.
    */
   public int indexOf(byte searchFor)
   {
      return indexOf(searchFor, 0);
   }

   /**
    * Returns the index within this byte array of the first occurrence of the
    * specified byte, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the byte in the larger byte
    *         array or -1 if not found.
    */
   public int indexOf(byte searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;

      // Is the starting position within the byte array?
      if (fromIndex >= arrayLen)
         return -1;

      for (int i = fromIndex; i < arrayLen; i++)
      {
         if (bArray[i] == searchFor)
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Returns the index within this larger byte array of the first occurrence of
    * the specified string.  Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the string in the byte array
    *         or -1 if not found.
    */
   public int indexOf(String searchFor)
   {
      return indexOf(searchFor, 0);
   }

   /**
    * Returns the index within this byte array of the first occurrence of the
    * specified string, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the string in the byte array
    *         or -1 if not found.
    */
   public int indexOf(String searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;
      int searchForLen = searchFor.length();

      // Starting point must be zero or higher.
      if (fromIndex < 0)
         return -1;
      // Is the starting position within the byte array?
      else if (fromIndex >= arrayLen)
         return -1;

      // Is the searchFor byte array larger than the byte array to be search?
      if (searchForLen > arrayLen)
         return -1;

      for (int i = fromIndex; i < arrayLen - searchForLen + 1; i++)
      {
         boolean found = true;
         for (int j = 0; j < searchForLen; j++)
         {
            if (bArray[i + j] != searchFor.charAt(j))
            {
               found = false;
               break;
            }
         }
         if (found)
            return i;
      }

      return -1;
   }

   /**
    * Returns the index within this byte array of the first occurrence of the
    * specified character.  Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the character in the byte
    *         array or -1 if not found.
    */
   public int indexOf(char searchFor)
   {
      return indexOf(searchFor, 0);
   }

   /**
    * Returns the index within this byte array of the first occurrence of the
    * specified character, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the character in the byte
    *         array or -1 if not found.
    */
   public int indexOf(char searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;

      // Is the starting position within the byte array?
      if (fromIndex >= arrayLen)
         return -1;

      for (int i = fromIndex; i < arrayLen; i++)
      {
         if (bArray[i] == searchFor)
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Returns true, if and only if, length is zero.
    *
    * @return true if length is zero, otherwise false.
    */
   public boolean isEmpty()
   {
      return (bArray.length == 0);
   }

   /**
    * Returns the index within this larger byte array of the last occurrence of
    * the specified byte array.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the byte array in the larger
    *         byte array or -1 if not found.
    */
   public int lastIndexOf(byte[] searchFor)
   {
      return lastIndexOf(searchFor, (bArray.length - searchFor.length - 1));
   }

   /**
    * Returns the index within this larger byte array of the last occurrence of
    * the specified byte array, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the byte array in the larger
    *         byte array or -1 if not found.
    */
   public int lastIndexOf(byte[] searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;
      int searchForLen = searchFor.length;

      // Is the starting position within the byte array?
      if (fromIndex >= arrayLen)
         return -1;

      // Is the searchFor byte array larger than the byte array to be search?
      if (searchForLen > arrayLen)
         return -1;

      for (int i = fromIndex; i >= 0; i--)
      {
         boolean found = true;
         for (int j = 0; j < searchForLen; j++)
         {
            if (bArray[i + j] != searchFor[j])
            {
               found = false;
               break;
            }
         }
         if (found)
            return i;
      }

      return -1;
   }

   /**
    * Returns the index within this byte array of the last occurrence of the
    * specified byte.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the byte in the larger byte
    *         array or -1 if not found.
    */
   public int lastIndexOf(byte searchFor)
   {
      return lastIndexOf(searchFor, (bArray.length - 1));
   }

   /**
    * Returns the index within this byte array of the last occurrence of the
    * specified byte, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the byte in the larger byte
    *         array or -1 if not found.
    */
   public int lastIndexOf(byte searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;

      // Starting point must be zero or higher.
      if (fromIndex < 0)
         return -1;
      // Is the starting position within the byte array?
      else if (fromIndex >= arrayLen)
         return -1;

      int i = Math.min(fromIndex, arrayLen - 1);

      for (; i >= 0; i--)
      {
         if (bArray[i] == searchFor)
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Returns the index within this larger byte array of the last occurrence of
    * the specified string.  Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the string in the byte array
    *         or -1 if not found.
    */
   public int lastIndexOf(String searchFor)
   {
      return lastIndexOf(searchFor, (bArray.length - 1));
   }

   /**
    * Returns the index within this byte array of the last occurrence of the
    * specified string, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the string in the byte array
    *         or -1 if not found.
    */
   public int lastIndexOf(String searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;
      int searchForLen = searchFor.length();

      // Is the starting position within the byte array?
      if (fromIndex >= arrayLen)
         return -1;

      // Is the searchFor byte array larger than the byte array to be search?
      if (searchForLen > arrayLen)
         return -1;

      for (int i = fromIndex; i >= 0; i--)
      {
         boolean found = true;
         for (int j = 0; j < searchForLen; j++)
         {
            if (bArray[i + j] != searchFor.charAt(j))
            {
               found = false;
               break;
            }
         }
         if (found)
            return i;
      }

      return -1;
   }

   /**
    * Returns the index within this byte array of the last occurrence of the
    * specified character.  Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @return the index of the first occurrence of the character in the byte
    *         array or -1 if not found.
    */
   public int lastIndexOf(char searchFor)
   {
      return lastIndexOf(searchFor, (bArray.length - 1));
   }

   /**
    * Returns the index within this byte array of the last occurrence of the
    * specified character, starting the search at a specified index.
    * Range is 0 to length - 1.
    *
    * @param searchFor value to be searched for
    * @param fromIndex starting position
    * @return the index of the first occurrence of the character in the byte
    *         array or -1 if not found.
    */
   public int lastIndexOf(char searchFor, int fromIndex)
   {
      int arrayLen = bArray.length;

      // Starting point must be zero or higher.
      if (fromIndex < 0)
         return -1;
      // Is the starting position within the byte array?
      else if (fromIndex >= arrayLen)
         return -1;

      int i = Math.min(fromIndex, arrayLen - 1);

      for (; i >= 0; i--)
      {
         if (bArray[i] == searchFor)
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Returns the length of the byte array.
    *
    * @return length of byte array
    */
   public int length()
   {
      return bArray.length;
   }

   /**
    * Replace all occurrences of oldChar in this byte array with newChar.
    * @param oldChar original character to search for and be replaced
    * @param newChar replacement character
    * @return returns the updated byte array
    */
   public byte[] replace(char oldChar, char newChar)
   {
      return replace((byte)oldChar, (byte)newChar);
   }

   /**
    * Replace all occurrences of oldByte in this byte array with newByte.
    * @param oldByte original byte to search for and be replaced
    * @param newByte replacement byte
    * @return returns the updated byte array
    */
   public byte[] replace(byte oldByte, byte newByte)
   {
      byte[] bOutput = new byte[bArray.length];

      for (int i = 0; i < bArray.length; i++)
      {
         if (bArray[i] == oldByte)
            bOutput[i] = newByte;
         else
            bOutput[i] = bArray[i];
      }

      return bOutput;
   }

   /**
    * Replace all occurrences of oldByteArray in this byte array with newByteArray.
    * @param oldByteArray original byte array to search for and be replaced
    * @param newByteArray replacement byte array
    * @return returns the updated byte array
    * @throws IndexOutOfBoundsException
    */
   public byte[] replace(byte[] oldByteArray, byte[] newByteArray) throws IndexOutOfBoundsException
   {
      byte[] bOutput = bArray.clone();

      // Is the oldByteArray byte array a different size than larger than newByteArray?
      if (oldByteArray.length != newByteArray.length)
         throw new IndexOutOfBoundsException("oldByteArray is a different size than newByteArray.");

      // Is the oldByteArray byte array larger than the byte array?
      if (oldByteArray.length > bArray.length)
         throw new IndexOutOfBoundsException("oldByteArray is larger than the byte array.");

      // Is the newByteArray byte array larger than the byte array?
      if (newByteArray.length > bArray.length)
         throw new IndexOutOfBoundsException("newByteArray is larger than the byte array.");

      int arrayLen = bArray.length;
      int oldByteArrayLen = oldByteArray.length;

      for (int i = 0; i < arrayLen - oldByteArrayLen + 1; i++)
      {
         boolean found = true;
         for (int j = 0; j < oldByteArrayLen; j++)
         {
            if (bArray[i + j] != oldByteArray[j])
            {
               found = false;
               break;
            }
         }
         if (found)
         {
            for (int j = 0; j < oldByteArrayLen; j++)
            {
               bOutput[i + j] = newByteArray[j];
            }
         }
      }

      return bOutput;
   }

   /**
    * Replace all occurrences of oldByteArray in this byte array with newByteArray.
    * @param oldByteArray original byte array to search for and be replaced
    * @param newByteArray replacement byte array
    * @throws IndexOutOfBoundsException
    */
   public void replaceAndResize(byte[] oldByteArray, byte[] newByteArray) throws IndexOutOfBoundsException
   {
      // Is the oldByteArray byte array larger than the byte array?
      if (oldByteArray.length > bArray.length)
         throw new IndexOutOfBoundsException("oldByteArray is larger than the byte array.");
      
      int pos = 0;;
      byte[] bOutput;

      while ((pos = indexOf(oldByteArray, pos)) != -1)
      {
         bOutput = new byte[bArray.length - oldByteArray.length + newByteArray.length];

         for (int i = 0; i < pos; i++)
         {
            bOutput[i] = bArray[i];
         }
         
         for (int i = pos, j=0; i < pos + newByteArray.length; i++, j++)
         {
            bOutput[i] = newByteArray[j];
         }
         
         int sOld = pos + oldByteArray.length;
         int sNew = pos + newByteArray.length;
         
         for (int  i=sNew, j=sOld; j < bArray.length; i++, j++)
         {
            bOutput[i] = bArray[j];
         }

         bArray = bOutput;
      }
   }

   /**
    * Resize the byte array to a new size.
    *
    * @param newSize the value for the new size of the byte array.
    * @return this byte array of a new size.
    */
   public byte[] resize(int newSize)
   {
      return Arrays.copyOf(bArray, newSize);
   }

   /**
    * Return an array of byte arrays. The original byte array is divided into
    * byte arrays by the specified delimiter.
    *
    * @param delimiter the character delimiter to be used as to split the byte array.
    * @return an array of byte arrays
    */
   public byte[][] split(char delimiter)
   {
      return split((byte) delimiter);
   }

   /**
    * Return an array of byte arrays. The original byte array is divided into
    * byte arrays by the specified delimiter.
    *
    * @param delimiter the byte delimiter to be used as to split the byte array.
    * @return an array of byte arrays
    */
   public byte[][] split(byte delimiter)
   {
      byte[][] bOutput = new byte[0][0];
      ArrayList<Integer> marker_Al = new ArrayList<Integer>();

      for (int i = 0; i < bArray.length; i++)
      {
         if (bArray[i] == delimiter)
            marker_Al.add(i);
      }

      if (marker_Al.size() > 0)
      {
         ArrayList<byte[]> bArray_Al = new ArrayList<byte[]>();
         int first = 0;
         int last = -1;
         byte[] bTemp;

         for (int i = 0; i < marker_Al.size(); i++)
         {
            first = last + 1;
            last = marker_Al.get(i);

            bTemp = Arrays.copyOfRange(bArray, first, last);
            bArray_Al.add(bTemp);
         }

         // add trailing data after last delimiter
         first = last + 1;
         last = bArray.length;
         bTemp = Arrays.copyOfRange(bArray, first, last);
         bArray_Al.add(bTemp);

         bOutput = new byte[bArray_Al.size()][0];
         for (int i = 0; i < bArray_Al.size(); i++)
         {
            bOutput[i] = bArray_Al.get(i);
         }
      }
      else
      { // no delimiters found.
         bOutput = new byte[1][0];
         bOutput[0] = bArray;
      }

      return bOutput;
   }

   /**
    * Tests if this byte array starts with the specified prefix.
    * @param prefix inputed byte array
    * @return true if the byte array starts with the inputed byte array, otherwise false
    */
   public boolean startsWith(byte[] prefix)
   {
      boolean flag = true;

      if (bArray.length < prefix.length)
         return false;

      for (int i = 0; i < prefix.length; i++)
      {
         if (bArray[i] != prefix[i])
         {
            flag = false;
            break;
         }
      }

      return flag;
   }

   /**
    * Tests if the sub byte array of this byte array beginning at the
    * specified index starts with the specified prefix.
    * @param prefix inputed byte array
    * @param offset offset within byte array to start the comparison at.
    * @return true if the byte array starts with the inputed byte array, otherwise false
    */
   public boolean startsWith(byte[] prefix, int offset)
   {
      boolean flag = true;

      if (bArray.length < (prefix.length + offset))
         return false;

      for (int i=offset, j=0; j < prefix.length; i++, j++)
      {
         if (bArray[i] != prefix[j])
         {
            flag = false;
            break;
         }
      }

      return flag;
   }

   /**
    * Return the entire byte array
    *
    * @return bArray
    */
   public byte[] toByteArray()
   {
      return bArray;
   }

   /**
    * Return the byte array as a character array
    *
    * @return an character array
    */
   public char[] toCharArray()
   {
      char[] cOutput = new char[bArray.length];

      for (int i = 0; i < bArray.length; i++)
         cOutput[i] = (char)bArray[i];

      return cOutput;
   }

   /**
    * Return the byte array as a hexadecimal representation.
    * @return byte array as a hex string
    */
   public String toHexArray()
   {
      char[] hexChars = new char[bArray.length * 2];
      for (int i = 0; i < bArray.length; i++)
      {
         int num = bArray[i] & 0xFF;
         hexChars[i * 2] = HEX_ARRAY[num >>> 4];
         hexChars[i * 2 + 1] = HEX_ARRAY[num & 0x0F];
      }

      return new String(hexChars);
   }

   /**
    * Return byte array as a string.
    *
    * @return aString
    */
   public String toString()
   {
      return new String(bArray);
   }
}
